#pragma once
#include <vectorclass.h>

// Macro for exporting DLL functions
#define DLL_EXPORT extern "C" __declspec(dllexport)

// Define for the mouse state
struct MouseState
{
	int PixelX, PixelY;
	int ScreenX, ScreenY;
	int buttonState;
} mouseState;

// Define for the keyboard state
struct KeyboardState
{
	int data[256];
} keyboardState;

// XOR shift random generator from Wikipedia, based on George Marsaglia
unsigned int state = 5345;
unsigned int XORRand()
{
	state ^= state << 13;
	state ^= state >> 17;
	state ^= state << 5;

	return state;
}

// 3 channel random number generator, generates 32 bytes for each of the RGB channels
// using 32 bit XOR shift generators
Vec8ui stateVCLR(7819, 9910, 7816, 3312, 8817, 3991, 6172, 8867);
Vec8ui stateVCLG(359663, 6482713, 4762640, 3021940, 268920, 5683340, 5340846, 5510688);
Vec8ui stateVCLB(1073667, 4242735, 2977588, 1946496, 194296, 3297146, 1983830, 3704747);

void XORRandVCL_RGB()
{
	stateVCLR ^= stateVCLR << 13;
	stateVCLG ^= stateVCLG << 5;
	stateVCLB ^= stateVCLB << 12;
	
	stateVCLR ^= stateVCLR >> 17;
	stateVCLG ^= stateVCLG >> 15;
	stateVCLB ^= stateVCLB >> 9;

	stateVCLR ^= stateVCLR << 5;	
	stateVCLG ^= stateVCLG << 17;
	stateVCLB ^= stateVCLB << 23;
}

// 16 bit color channel
class Channel16
{
public:
	unsigned short* data;
	int width, height;

	// Construct new channel
	Channel16(int width, int height)
	{
		this->width = width;
		this->height = height;

		data = (unsigned short*)_aligned_malloc(
			(sizeof(unsigned short) * width * height) +
			512,	// Extra bytes so we don't worry about SIMD going over bounds
			256);

		// Clear it
		for (int i = 0; i < width * height; i++)
			data[i] = 0;
	}

	// Delete this channel
	~Channel16()
	{
		_aligned_free(data);

		data = 0;
	}
};

// Collection of 3 16 bit channels, R, G and B
class RGB16bpc
{
public:
	Channel16* r;
	Channel16* g;
	Channel16* b;

	int width, height;

	int Count()
	{
		return width * height;
	}

	RGB16bpc(int width, int height)
	{
		this->width = width;
		this->height = height;

		r = new Channel16(width, height);
		g = new Channel16(width, height);
		b = new Channel16(width, height);
	}

	~RGB16bpc()
	{
		delete r;
		delete g;
		delete b;
	}
};

// Clears RGB channels to black
DLL_EXPORT void Clear(void* rgb, unsigned short r, unsigned short g, unsigned short b)
{
	RGB16bpc* x = (RGB16bpc*)rgb;

	for (int i = 0; i < x->width * x->height; i++)
	{
		x->r->data[i] = r;
		x->g->data[i] = g;
		x->b->data[i] = b;
	}
}

// Allocate a 16 bit channel of the specified size
DLL_EXPORT void* AllocateRGB16bpc(int width, int height)
{
	RGB16bpc* rgb = new RGB16bpc(width, height);

	return rgb;
}

// Function to update the input, called from the UI when input changes
DLL_EXPORT void UpdateInput(void* mouse, void* keyboard)
{
	// Update the mouse state
	mouseState.PixelX = ((MouseState*)mouse)->PixelX;
	mouseState.PixelY = ((MouseState*)mouse)->PixelY;
	mouseState.ScreenX = ((MouseState*)mouse)->ScreenX;
	mouseState.ScreenY = ((MouseState*)mouse)->ScreenY;
	mouseState.buttonState = ((MouseState*)mouse)->buttonState;

	// Do you have to copy? Maybe, you cannot be sure that the 'fixed' array
	// from C# won't change when that code block ends
	for (int i = 0; i < 256; i++)
	{
		keyboardState.data[i] = ((KeyboardState*)keyboard)->data[i];
	}
}

// Pack the 16 bit channels to 8 bits for display, original algo without SIMD
DLL_EXPORT void RGB16bpcToBMP_CPP(void* dest, void* src, bool dithering)
{
	int height = ((RGB16bpc*)src)->height;
	int width = ((RGB16bpc*)src)->width;

	unsigned short* r = ((RGB16bpc*)src)->r->data;
	unsigned short* g = ((RGB16bpc*)src)->g->data;
	unsigned short* b = ((RGB16bpc*)src)->b->data;
	int count = ((RGB16bpc*)src)->width * ((RGB16bpc*)src)->height;
	unsigned int* d = (unsigned int*)dest;
	if(dithering)
	{
		for (int i = 0; i < count; i++)
		{
			unsigned int R = (r[i] >> 8) & 0xff;
			unsigned int G = (g[i] >> 8) & 0xff;
			unsigned int B = (b[i] >> 8) & 0xff;

			// Dither
			if ((XORRand() & 0xff) < (r[i] & 0xff) && R != 255)
				R++;
			if ((XORRand() & 0xff) < (g[i] & 0xff) && G != 255)
				G++;
			if ((XORRand() & 0xff) < (b[i] & 0xff) && B != 255)
				B++;

			unsigned int col = 0xff000000;
			col |= B << 0;
			col |= G << 8;
			col |= R << 16;

			d[i] = col;
		}
	}
	else
	{
		for (int i = 0; i < count; i++)
		{
			unsigned int R = (r[i] >> 8) & 0xff;
			unsigned int G = (g[i] >> 8) & 0xff;
			unsigned int B = (b[i] >> 8) & 0xff;

			unsigned int col = 0xff000000;
			col |= B << 0;
			col |= G << 8;
			col |= R << 16;

			d[i] = col;
		}
	}
}

DLL_EXPORT void RGB16bpcToBMP_VCL(void* dest, void* src, bool dithering)
{
	unsigned short* r = ((RGB16bpc*)src)->r->data;
	unsigned short* g = ((RGB16bpc*)src)->g->data;
	unsigned short* b = ((RGB16bpc*)src)->b->data;

	RGB16bpc* source = (RGB16bpc*)src;
	int count = source->Count();
	unsigned int* d = (unsigned int*)dest;

	Vec16us R1, R2;		// 16 red values as Shorts
	Vec32uc R_Chance;	// 32 bytes representing the chances of dithering
	Vec32uc R;			// The Red channel as bytes

	Vec16us G1, G2;		// 16 green values as shorts
	Vec32uc G_Chance;	// 32 bytes representing the chances of dithering
	Vec32uc G;			// The green channel as bytes

	Vec16us B1, B2;		// 16 blue values as shorts
	Vec32uc B_Chance;	// 32 bytes representing the chances of dithering
	Vec32uc B;			// The blue channel as bytes

	// Vector of 100% opaque, for when combining channels
	Vec8ui _ALPHA_OPAQUE = 0xff000000;

	// Count how many residuals there are, pixels we have to use scalar to compute
	int resid = count & 31;

	// The current pixel, or start of the SIMD pixels we're operating on
	int pixelIndex = 0;
	if (dithering)
	{
		for (; pixelIndex < count; pixelIndex += 32)
		{
			// Read 32 16 bit values
			R1.load(r + pixelIndex);
			R2.load(r + pixelIndex + 16);
			G1.load(g + pixelIndex);
			G2.load(g + pixelIndex + 16);
			B1.load(b + pixelIndex);
			B2.load(b + pixelIndex + 16);

			// Chance of incrementing is 1/X, where X is the lower 8 bits of each 16 bit word
			R_Chance = compress(R1, R2);
			G_Chance = compress(G1, G2);
			B_Chance = compress(B1, B2);

			// The main red component is the upper 8 bits of each 16 bit word
			R = compress(R1 >> 8, R2 >> 8);
			G = compress(G1 >> 8, G2 >> 8);
			B = compress(B1 >> 8, B2 >> 8);

			// Compute which bytes would increment, will result in 0xff for +1, and 0x00 for not
			XORRandVCL_RGB();
			R_Chance = (reinterpret_i(stateVCLR) < R_Chance);
			G_Chance = (reinterpret_i(stateVCLG) < G_Chance);
			B_Chance = (reinterpret_i(stateVCLB) < B_Chance);

			// AND with 1 so these increments are +1 or 0
			R_Chance = R_Chance & 1;
			G_Chance = G_Chance & 1;
			B_Chance = B_Chance & 1;

			// Add increments, saturating at 255
			R = add_saturated(R, R_Chance);
			G = add_saturated(G, G_Chance);
			B = add_saturated(B, B_Chance);

			// Combine into 32 uint RGBA values
			Vec16us r1 = extend_low(R);
			Vec16us r2 = extend_high(R);
			Vec8ui r3 = extend_low(r1);
			Vec8ui r4 = extend_high(r1);
			Vec8ui r5 = extend_low(r2);
			Vec8ui r6 = extend_high(r2);

			Vec16us g1 = extend_low(G);
			Vec16us g2 = extend_high(G);
			Vec8ui g3 = extend_low(g1);
			Vec8ui g4 = extend_high(g1);
			Vec8ui g5 = extend_low(g2);
			Vec8ui g6 = extend_high(g2);

			Vec16us b1 = extend_low(B);
			Vec16us b2 = extend_high(B);
			Vec8ui b3 = extend_low(b1);
			Vec8ui b4 = extend_high(b1);
			Vec8ui b5 = extend_low(b2);
			Vec8ui b6 = extend_high(b2);

			Vec8ui combined1 = _ALPHA_OPAQUE | (r3 << 16) | (g3 << 8) | b3;
			Vec8ui combined2 = _ALPHA_OPAQUE | (r4 << 16) | (g4 << 8) | b4;
			Vec8ui combined3 = _ALPHA_OPAQUE | (r5 << 16) | (g5 << 8) | b5;
			Vec8ui combined4 = _ALPHA_OPAQUE | (r6 << 16) | (g6 << 8) | b6;

			combined1.store(d + pixelIndex);
			combined2.store(d + pixelIndex + 8);
			combined3.store(d + pixelIndex + 16);
			combined4.store(d + pixelIndex + 24);
		}

		// The scalar loop handles any additional residual pixels that the SIMD loop doesn't compute
		for (int i = pixelIndex; i < pixelIndex + resid; i++)
		{
			unsigned int R = (r[i] >> 8) & 0xff;
			unsigned int G = (g[i] >> 8) & 0xff;
			unsigned int B = (b[i] >> 8) & 0xff;

			// Dither
			if ((XORRand() & 0xff) < (r[i] & 0xff) && R != 255)
				R++;
			if ((XORRand() & 0xff) < (g[i] & 0xff) && G != 255)
				G++;
			if ((XORRand() & 0xff) < (b[i] & 0xff) && B != 255)
				B++;

			unsigned int col = 0xff000000;
			col |= B << 0;
			col |= G << 8;
			col |= R << 16;

			d[i] = col;
		}
	}

	// If dithering is turned off:
	else
	{
		for (; pixelIndex < count; pixelIndex += 32)
		{
			// Read 32 16 bit values
			R1.load(r + pixelIndex);
			R2.load(r + pixelIndex + 16);
			G1.load(g + pixelIndex);
			G2.load(g + pixelIndex + 16);
			B1.load(b + pixelIndex);
			B2.load(b + pixelIndex + 16);

			// The main red component is the upper 8 bits of each 16 bit word
			// There is no dithering, so ignore the lower byte
			R = compress(R1 >> 8, R2 >> 8);
			G = compress(G1 >> 8, G2 >> 8);
			B = compress(B1 >> 8, B2 >> 8);

			// Combine into 32 uint RGBA values
			Vec16us r1 = extend_low(R);
			Vec16us r2 = extend_high(R);
			Vec8ui r3 = extend_low(r1);
			Vec8ui r4 = extend_high(r1);
			Vec8ui r5 = extend_low(r2);
			Vec8ui r6 = extend_high(r2);

			Vec16us g1 = extend_low(G);
			Vec16us g2 = extend_high(G);
			Vec8ui g3 = extend_low(g1);
			Vec8ui g4 = extend_high(g1);
			Vec8ui g5 = extend_low(g2);
			Vec8ui g6 = extend_high(g2);

			Vec16us b1 = extend_low(B);
			Vec16us b2 = extend_high(B);
			Vec8ui b3 = extend_low(b1);
			Vec8ui b4 = extend_high(b1);
			Vec8ui b5 = extend_low(b2);
			Vec8ui b6 = extend_high(b2);

			Vec8ui combined1 = _ALPHA_OPAQUE | (r3 << 16) | (g3 << 8) | b3;
			Vec8ui combined2 = _ALPHA_OPAQUE | (r4 << 16) | (g4 << 8) | b4;
			Vec8ui combined3 = _ALPHA_OPAQUE | (r5 << 16) | (g5 << 8) | b5;
			Vec8ui combined4 = _ALPHA_OPAQUE | (r6 << 16) | (g6 << 8) | b6;

			combined1.store(d + pixelIndex);
			combined2.store(d + pixelIndex + 8);
			combined3.store(d + pixelIndex + 16);
			combined4.store(d + pixelIndex + 24);
		}

		// The scalar loop handles any additional residual pixels that the SIMD loop doesn't compute
		for (int i = pixelIndex; i < pixelIndex + resid; i++)
		{
			unsigned int R = (r[i] >> 8) & 0xff;
			unsigned int G = (g[i] >> 8) & 0xff;
			unsigned int B = (b[i] >> 8) & 0xff;

			unsigned int col = 0xff000000;
			col |= B << 0;
			col |= G << 8;
			col |= R << 16;

			d[i] = col;
		}
	}
}

// Render the 16 bits per channel down to 8 for display
DLL_EXPORT void RGB16bpcToBMP(void* dest, void* src, bool dithering, bool vcl)
{
	if (vcl)
		RGB16bpcToBMP_VCL(dest, src, dithering);
	else
		RGB16bpcToBMP_CPP(dest, src, dithering);
}

// Deallocate a collection of RGB channels
DLL_EXPORT void Deallocate(void* rgb16bpc)
{
	RGB16bpc* rgb = (RGB16bpc*)rgb16bpc;

	delete rgb;
}