﻿// Special note for Patreons:
// This code does not handle residuals in the vectorised fractals
// in resolutions with pixel counts not evently divisible by 8.

#include <iostream>
#include <vectorclass.h>
#include <vectormath_trig.h>
#include <vectormath_hyp.h>
#include <vectormath_exp.h>
#include <complex>
#include <ctime>
#include <thread>
#include "RenderBaby.h"

#include "JuliaGradient.h"
#include "SkewedGradient.h"
#include "KaleidoscopeGradient.h"
#include "RobotsGradient.h"
#include "VampireGradient.h"
#include "WaterGradient.h"

// Pointer to the pixel data as 16 bits per channel
RGB16bpc* rgb;

// Width and height of the screen in pixels, emulated
int width, height;

// Scale and time scale for displaying and cycling the fractals
const float SCALE = 0.01f;
const float TIME_SCALE = 10.0f;

// Various settable parameters changeable with keyboard
float cTime = 0.0f;
Vec8f vCTime = 0.0f;
float Zoom = 1.0f;
const float MIN_ZOOM = 0.1f;
const float PAN_SPEED = 500.0f;
float PanX = 0.0f;
float PanY = 0.0f;
float* currentGradient = juliaGradient;

int FractalIndex = 0;	// The selected fractal's index, 0 to 5; 0 is Julia, 5 is Vampire

// Escape magnitude and the repetitions before giving up
float escapeRoot = 1000000.0f;
float reps = 256.0f;	// 256 because there's 256 colors in the gradients

// Maximum value of an unsigned short integer as float vec
Vec8f shortMax = (float)0x0000ffff;

// You can truncate and compress to shorts in SIMD instead of arr[x] etc.
void PlotVec8f(int x, int y, Vec8f& r1, Vec8f& g1, Vec8f& b1)
{
	/*
	// This original code plots R, G and B as normal, the updated code
	// reads values from a LUT so the colors can be more appealing.
	// Read 8 floats from Red, multiply by 65536
	Vec8i i = truncatei(r1 * shortMax);
	// Compress to just the low byte since the gradient has only these values
	i = i >> 8;

	// Multiply the indices by 3 for the RGB channels
	i *= 3;

	// Add 0 (nothing) for the Red channel
	i += 0;

	// Lookup the colors in the table
	Vec8f gradientLookup = lookup<8>(i, (float*)juliaGradient);

	// Convert to short again
	i = truncatei(gradientLookup * shortMax);
	Vec8s s = compress(i);
	s.store(&rgb->r->data[x + y * width]);

	i = truncatei(g1 * shortMax);
	s = compress(i);
	s.store(&rgb->g->data[x + y * width]);

	i = truncatei(b1 * shortMax);
	s = compress(i);
	s.store(&rgb->b->data[x + y * width]);*/

	// New code:

	// Red
	// Read 8 floats from Red, mul by 256 to get the table index, add 0 for Red channel
	Vec8i i = truncatei(r1 * 256) * 3 + 0;
	
	// Lookup the colors in the table
	// Note: Template parameter is the size of the lookup table
	Vec8f gradientLookup = lookup<256*3>(i, currentGradient);

	// Convert to shorts since 16 bits per channel
	Vec8s s = compress(truncatei(gradientLookup * shortMax));
	
	// Store
	s.store(&rgb->r->data[x + y * width]);

	// Green
	i = truncatei(g1 * 256) * 3 + 1;
	gradientLookup = lookup<256*3>(i, currentGradient);
	i = truncatei(gradientLookup * shortMax);
	s = compress(i);
	s.store(&rgb->g->data[x + y * width]);

	// Blue
	i = truncatei(b1 * 256) * 3 + 2;
	gradientLookup = lookup<256*3>(i, currentGradient);
	i = truncatei(gradientLookup * shortMax);
	s = compress(i);
	s.store(&rgb->b->data[x + y * width]);
}

// Called once when the program begins executing
DLL_EXPORT void Init(int* widthHeight)
{
	// Emulated resolution
	width = 1920;
	height = 1280;
	//width = 800;
	//height = 600;

	widthHeight[0] = width;
	widthHeight[1] = height;
}

// Called once when the program is shutting down, release resources allocated in init()
DLL_EXPORT void Shutdown()
{

}

void Plot(int x, int y, unsigned short r, unsigned short g, unsigned short b)
{
	if (x < 0 || x >= width || y < 0 || y >= height)
		return;

	rgb->r->data[x + y * width] = r;
	rgb->g->data[x + y * width] = g;
	rgb->b->data[x + y * width] = b;
}

// Small, simple complex number struct for C++ computations
struct Complex
{
	float r, i;

	Complex(float r, float i)
	{
		this->r = r;
		this->i = i;
	}
};

// Add complex numbers a and b, store sum in a
Complex Add(Complex& a, Complex& b)
{
	Complex res(a.r + b.r, a.i + b.i);
	return res;
}

// Sub complex number b from a, store result in a
Complex Sub(Complex& a, Complex& b)
{
	Complex res(a.r - b.r, a.i - b.i);
	return res;
}

// SIMD complex structure of 8 floating point complex numbers
struct Comp8f
{
	Vec8f r;
	Vec8f i;

	Comp8f(Vec8f real, Vec8f imag)
	{
		r = real;
		i = imag;
	}

	Comp8f()
	{
		r = i = 0;
	}
};

// Add 8 complex numbers in float vectors, store sum in a
Comp8f Add(Comp8f& a, Comp8f& b)
{
	Comp8f res;
	res.r = a.r + b.r;
	res.i = a.i + b.i;
	return res;
}

// Subtract 8 complex numbers in bector B from those in a, store result in a
Comp8f Sub(Comp8f& a, Comp8f& b)
{
	Comp8f res;
	res.r = a.r - b.r;
	res.i = a.i - b.i;
	return res;
}

// Multiplication function for C++
Complex Mul(Complex& a, Complex& b)
{
	// By changing the steps for multiplication, and introducing
	// various arbitrary functions, we can create interesting
	// variations on the standard Julia Set.

	// resulting complex number
	Complex res(0, 0);
	switch (FractalIndex)
	{
		case 0:	// Normal
			res.r = a.r * b.r - b.i * a.i;
			res.i = a.r*b.i+a.i*b.r;
			break;
		case 1:	// Skewed
			res.r = a.r * b.r - b.i * a.i;
			res.i = a.r*b.i;
			break;
		case 2:	// Water
			res.r = a.r * b.r - b.i * a.i;
			res.i = sin(a.r*b.i) + cos(a.i*b.r);
			break;
		case 3: // Robots
			res.r = a.r * b.r - b.i * a.i;
			res.i = (a.r*a.i + b.r * b.i)*(a.r*a.i + b.r * b.i);
			break;
		case 4: // Kaleidoscope
			res.r = a.r * b.r - b.i * a.i;
			res.i = pow(sin(a.r*a.i + b.r * b.i), 2);
			break;

		case 5: // Vampire :O
			res.r = a.r * b.r - b.i * a.i;
			res.i = a.r*pow(b.i, 2) * sin(pow(a.i, 2)/b.r);
			break;
	};

	return res;
}

// Multiplication function for SIMD complex vectors
Comp8f Mul(Comp8f& a, Comp8f& b)
{
	// Resulting vector
	Comp8f res;

	switch (FractalIndex)
	{
	case 0:
		// Normal
		res.r = a.r*b.r - b.i*a.i;
		res.i = a.r*b.i + a.i*b.r;

		break;
	case 1: 
		// Skewed
		res.r = a.r * b.r - b.i * a.i;
		res.i = a.r*b.i;
		break;
	case 2:	// Water
		res.r = a.r * b.r - b.i * a.i;
		res.i = sin(a.r*b.i) + cos(a.i*b.r);
		break;
	case 3:	// Robots
		res.r = a.r * b.r - b.i * a.i;
		res.i = (a.r*a.i + b.r * b.i)*(a.r*a.i + b.r * b.i);
		break;
	case 4:	// Kaleidoscope
		res.r = a.r * b.r - b.i * a.i;
		res.i = pow(sin(a.r*a.i + b.r * b.i), 2);
		
		break;
	case 5: // Vampire :O
		res.r = a.r * b.r - b.i * a.i;
		res.i = a.r*pow(b.i, 2) * sin(pow(a.i, 2)/b.r);
		break;
	};

	return res;
}

// Function to check if C++ number has escaped
bool EscapeMagnitudeCheck(Complex& a, float magnitudeRoot)
{
	return ((a.r*a.r) + (a.i * a.i)) >= magnitudeRoot;
}

// Function to check if the SIMD complex numbers escape, returns a Boolean vector
Vec8fb EscapeMagnitudeCheck8f(Comp8f& a, Vec8f& magnitudeRoot)
{
	return ((a.r*a.r) + (a.i * a.i)) >= magnitudeRoot;
}

// Respond to the user input
void Update(double timeTotal, double timeDelta)
{
	// Cycle back and forth through fractals
	if (keyboardState.data['P'] == 1)
	{
		vCTime += (float)timeDelta * TIME_SCALE;
		cTime += timeDelta * TIME_SCALE;
	}
	if (keyboardState.data['L'] == 1)
	{
		vCTime -= (float)timeDelta * TIME_SCALE;
		cTime -= timeDelta * TIME_SCALE;
	}

	// Zoom
	if (keyboardState.data['Z'] == 1)
	{
		Zoom += 0.01;
	}
	if (keyboardState.data['X'] == 1)
	{
		Zoom -= 0.01;

		if (Zoom < MIN_ZOOM)
			Zoom = MIN_ZOOM;
	}

	// Pan
	if (keyboardState.data['A'] == 1)
		PanX -= (PAN_SPEED * timeDelta) / Zoom;
	if (keyboardState.data['D'] == 1)
		PanX += (PAN_SPEED * timeDelta) / Zoom;
	if (keyboardState.data['W'] == 1)
		PanY -= (PAN_SPEED * timeDelta) / Zoom;
	if (keyboardState.data['S'] == 1)
		PanY += (PAN_SPEED * timeDelta) / Zoom;

	// Changing the fractals
	if (keyboardState.data['1'] == 1)
	{
		FractalIndex = 0;
		currentGradient = juliaGradient;
	}
	if (keyboardState.data['2'] == 1)
	{
		FractalIndex = 1;
		currentGradient = skewedGradient;
	}
	if (keyboardState.data['3'] == 1)
	{
		FractalIndex = 2;
		currentGradient = waterGradient;
	}
	if (keyboardState.data['4'] == 1)
	{
		FractalIndex = 3;
		currentGradient = robotsGradient;
	}	
	if (keyboardState.data['5'] == 1)
	{
		FractalIndex = 4;
		currentGradient = kaleidoscopeGradient;
	}
	if (keyboardState.data['6'] == 1)
	{
		FractalIndex = 5;
		currentGradient = vampireGradient;
	}
	// Reset
	if (keyboardState.data['R'] == 1)
	{
		Zoom = 1;
		PanX = PanY = 0;
		cTime = 0.0;
		vCTime = 0.0;
	}
}

// Render fractals using C++
DLL_EXPORT void Render16bpc_CPP(void* A000, double timeTotal, double timeDelta)
{
	// Update, read any inputs, etc.
	Update(timeTotal, timeDelta);

	// Pointer to pixel data
	rgb = (RGB16bpc*)A000;

	// The C value, or additive value in the fractal equation
	Complex CR(sin(cTime/31.0f), cos(cTime/97.0f));

	// The escape value is the escape root squared
	float escape = escapeRoot * escapeRoot;

	for (int y = 0; y < height; y++)
	{
		for (int x = 0; x < width; x++)
		{
			// Initial values
			float iv1 = ((float)((x+PanX)*Zoom) * 0.0025f)-2.5f;
			float iv2 = ((float)((y+PanY)*Zoom) * 0.0025f)-1.5f;
			
			Complex compVal(iv1, iv2);

			float repr = 0.0f;
			for (; repr < reps; repr++)
			{
				compVal = Mul(compVal, compVal);
				compVal = Add(compVal, CR);

				if (EscapeMagnitudeCheck(compVal, escape))
					break;
			}

			float R = 0.0f;
			if (EscapeMagnitudeCheck(compVal, escape))
				R = repr/reps;

			// Make it brighter, square and square
			// This was old code before the gradients were employed
			//R = sqrt(sqrt(R));
			//Plot(x, y, (unsigned short)(R * (float)0xffff),	0, 0);

			// Gradient lookup to get colors and render
			unsigned short rr = currentGradient[(int)(R * 256)*3+0] * 0xffff;
			unsigned short gg = currentGradient[(int)(R * 256)*3+1] * 0xffff;
			unsigned short bb = currentGradient[(int)(R * 256)*3+2] * 0xffff;
			Plot(x, y, rr, gg, bb);
		}
	}
}

// Render fractals using multi-threaded VCL
DLL_EXPORT void Render16bpc_VCL_MultiThread(void* A000, double timeTotal, double timeDelta, int threadId, int threadCount)
{
	// Pointer to pixel data
	rgb = (RGB16bpc*)A000;

	// Sine waves controlled by P and L keys to render different variations of the
	// current fractal
	Comp8f CR(sin(vCTime/31.0f), cos(vCTime/97.0f));

	// Constants for computing
	Vec8f ZERO = 0.0f;
	Vec8f escapeRootVec = escapeRoot;

	// The escape value is the escape root squared
	Vec8f escape = escapeRootVec * escapeRootVec;
	Vec8fb escapedSoFar;

	Vec8f escapedIteration = 0.0f;

	Vec8f EIGHTS = 8.0f;
	Vec8fb escapedThisRound;
	Vec8f escaped;
	Vec8f escapedRound;
	Vec8f currentRound;

	// Variables to allocate work loads for each thread
	int workLoad = height / threadCount;
	int startHeight = workLoad * threadId;
	Vec8f Zoom8f = Zoom;
	Vec8f PanX8f = PanX;
	Vec8f PanY8f = PanY;

	for (int y = startHeight; y < startHeight + workLoad; y++)
	{
		Vec8f XX(0, 1, 2, 3, 4, 5, 6, 7);
		Vec8f YY = y;
		
		// For Mandelbrot
		//Complex8f Z(0.0f, 0.0f);

		for (int x = 0; x < width; x += 8)
		{
			Comp8f compVal(((XX+PanX)*Zoom8f * 0.0025f)-2.5f, ((YY+PanY)*Zoom8f * 0.0025f)-1.5f);
			
			// For Mandelbrot
			//Complex8f mandelC(((XX+PanX)*Zoom8f * 0.0025f)-2.5f, ((YY+PanY)*Zoom8f * 0.0025f)-1.5f);

			escapedRound = 0.0f;
			currentRound = 0.0;
			escapedSoFar = false;
			escaped = 0.0f;
			int iteration;
			
			// For Mandelbrot
			//Z.r = 0.0f;
			//Z.i = 0.0f;

			for (iteration = 0; iteration < reps; iteration++)
			{
				compVal = Mul(compVal, compVal);
				compVal = Add(compVal, CR);

				// For Mandelbrot
				//Z = Mul(Z, Z);
				//Z = Add(Z, mandelC);

				// Who is escaped this round?
				escapedThisRound = EscapeMagnitudeCheck8f(compVal, escape);
				
				// For Mandelbrot
				//escapedThisRound = EscapeMagnitudeCheck8f(Z, escape);

				// Who is escaped so far?
				escaped = currentRound & escapedThisRound;

				// Who of them escaped this round? (this the new escapees, escaped^oldEscapes
				escaped = escaped & (escapedSoFar ^ escapedThisRound);

				// OR those that escaped this round with all the escaped so far
				escapedSoFar |= escapedThisRound;

				// Record the escape rounds for any that did escape this round
				escapedRound |= escaped;

				// Next round
				currentRound++;

				// If all have escaped, break
				if (horizontal_and(escapedSoFar))
					break;
			}

			// Doesn't really matter where we store the computed val, reuse .r
			compVal.r = escapedRound / reps;

			// Red is the gradient value, it's 0.0 to 1.0 at the moment
			PlotVec8f(x, y, compVal.r, compVal.r, compVal.r);

			// Increment the 8 X for loop counters
			XX = XX + EIGHTS;
		}
	}
}

// Creates threads and calls the VCL rendering
DLL_EXPORT void Render16bpc_VCL(void* A000, double timeTotal, double timeDelta)
{
	Update(timeTotal, timeDelta);

	std::thread* t[16];

	int threadCount = 8;

	for (int i = 0; i < threadCount; i++)
		t[i] = new std::thread(Render16bpc_VCL_MultiThread, A000, timeTotal, timeDelta, i, threadCount);

	for (int i = 0; i < threadCount; i++)
	{
		t[i]->join();
		delete t[i];
	}
}

// Rendering function, calls either C++ ro VCL fractal rendering functions
DLL_EXPORT void Render16bpc(void* A000, double timeTotal, double timeDelta, bool vcl)
{
	if (vcl)
		Render16bpc_VCL(A000, timeTotal, timeDelta);
	else
		Render16bpc_CPP(A000, timeTotal, timeDelta);
}

// For debugging and testing outside of DLL, compile to EXE and use main.
// I do not know what this code tests, it was whatever was last debugged
// before switching to DLL and compiling for the video.
int main()
{
	Vec8f t(0.0f, 0.5f, 0.25f, 0.1f, 0.0f, 0.5f, 0.25f, 0.1f);
	PlotVec8f(0, 0, t, t, t);

	int width = 1920;
	int height = 1080;

	RGB16bpc* v = (RGB16bpc*)AllocateRGB16bpc(width, height);
	unsigned char* dest = new unsigned char[width * height*4];
	
	for (int i = 0; i < v->Count(); i++)
	{
		v->r->data[i] = rand();
		v->g->data[i] = rand();
		v->b->data[i] = rand();
	}

	for (int i = 0; i < 10; i++)
	{
		long startTime = clock();

		for(int r = 0; r < 100; r++)
			RGB16bpcToBMP(dest, v, true, true);

		// 1920x1080
		// 139 vs 2687
		// 157 with 3 calls to the XORRandVCL
		long finishTime = clock();

		std::cout<<"ti: "<< (finishTime - startTime) << std::endl;
	}
	std::cin.get();

	delete[] dest;

	Deallocate(v);

	return 0;
}