﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Drawing.Imaging;
using System.IO;

namespace RenderLab2RenderBaby
{
	public unsafe class Surface
	{
		/// <summary>
		/// Context for graphics
		/// </summary>
		private BufferedGraphicsContext context = null;
		
		/// <summary>
		/// Buffered graphics object
		/// </summary>
		private BufferedGraphics grafx = null;
		
		/// <summary>
		/// The control this graphics renders the bitmap to
		/// </summary>
		private Control control = null;

		/// <summary>
		/// The bitmap the use draws on
		/// </summary>
		private Bitmap backBuffer = null;

		int width, height;	// Width and height of the control
		int resX, resY;     // The resolution of the emulated graphics, dimensions of the bitmap
		uint* A000;			// Pointer to the screen data

		public int Width
		{
			get
			{
				return resX;
			}
		}

		public int Height
		{
			get
			{
				return resY;
			}
		}

		/// <summary>
		/// Aspect ratio of the resolution
		/// </summary>
		double resAspect;

		/// <summary>
		/// Aspect ration of the control
		/// </summary>
		double controlAspect;

		/// <summary>
		/// Variables for tracking times
		/// </summary>
		DateTime initTime;
		DateTime lastTime;
		DateTime curTime;
		DateTime pluginLastTime;

		/// <summary>
		/// Bools for dithering and vcl
		/// </summary>
		public bool dithering = false;
		public bool vcl = false;

		[DllImport("RenderBaby.dll", CharSet = CharSet.Unicode)]
			public static extern void* AllocateRGB16bpc(
				int width,
				int height
				);

		/// <summary>
		/// Import to render 16 bit channels down to 2
		/// </summary>
		/// <param name="dest"></param>
		/// <param name="src"></param>
		/// <param name="dithering"></param>
		/// <param name="vcl"></param>
		[DllImport("RenderBaby.dll", CharSet = CharSet.Unicode)]
		public static extern void RGB16bpcToBMP(void* dest, void* src, bool dithering, bool vcl);

		/// <summary>
		/// Import for the plugin function, function that renders the fractals
		/// </summary>
		/// <param name="rgb16bpc"></param>
		/// <param name="timeTotal"></param>
		/// <param name="timeDelta"></param>
		/// <param name="vcl"></param>
		[DllImport("RenderBaby.dll", CharSet = CharSet.Unicode)]
		public static extern void Render16bpc(void* rgb16bpc, double timeTotal, double timeDelta, bool vcl);

		/// <summary>
		/// Import to deallocate resources
		/// </summary>
		/// <param name="rgb16bpc"></param>
		[DllImport("RenderBaby.dll", CharSet = CharSet.Unicode)]
		public static extern void Deallocate(void* rgb16bpc);

		/// <summary>
		/// Pointer to 16 bits per channel pixel data
		/// </summary>
		void* rgb16bpc = null;

		/// <summary>
		/// Create a new Surface with the control specified, and emulated resolution
		/// </summary>
		/// <param name="control">The control to render to</param>
		/// <param name="horizontal_resolution">The emulated horizontal resolution in pixels</param>
		/// <param name="vertical_resolution">The emulated vertical resolution in pixels</param>
		public Surface(Control control, int horizontal_resolution, int vertical_resolution)
		{
			this.control = control;
			this.resX = horizontal_resolution;
			this.resY = vertical_resolution;

			Reset();

			initTime = DateTime.Now;
			curTime = initTime;
			lastTime = initTime;
		}

		/// <summary>
		/// Recomputes the sizes of the surface and control, allocates the memory
		/// </summary>
		public void Reset()
		{
			// Dipose if there's allocated RAM
			Dispose();

			// Create double buffered graphics
			context = BufferedGraphicsManager.Current;

			context.MaximumBuffer = new Size(control.Width, control.Height);

			grafx = context.Allocate(control.CreateGraphics(),
				 new Rectangle(0, 0, control.Width, control.Height));

			grafx.Graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.NearestNeighbor;
			
			// Set the size of the output
			width = control.Width;

			height = control.Height;

			resAspect = (double)resX / (double)resY;

			controlAspect = (double)control.Width / (double)control.Height;

			// Allocate our drawing bitmap
			backBuffer = new Bitmap(resX, resY);

			// Allocate the 16 bit per channel buffer
			rgb16bpc = AllocateRGB16bpc(resX, resY);
		}

		/// <summary>
		/// Switch 16 bits per channel on and off
		/// </summary>
		public void ToggleDithering()
		{
			dithering = !dithering;
		}

		/// <summary>
		/// Switch VCL on and off
		/// </summary>
		public void ToggleVCL()
		{
			vcl = !vcl;
		}

		/// <summary>
		/// Disposes resources for this surface
		/// </summary>
		public void Dispose()
		{
			// Dispose of any resources if they exist
			if (grafx != null)
			{
				grafx.Dispose();
				grafx = null;
			}
			if (backBuffer != null)
			{
				backBuffer.Dispose();
				backBuffer = null;
			}
			if (rgb16bpc != null)
			{
				Deallocate(rgb16bpc);
				rgb16bpc = null;
			}
		}

		public void Redraw()
		{
			// Lock the bitmap, it is 32bpp ARGB, 8 bits per channel
			BitmapData bmpData = backBuffer.LockBits(
				new Rectangle(0, 0,
				backBuffer.Width, backBuffer.Height),
				ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);

			// Get the pointer to the pixel data
			A000 = (uint*)bmpData.Scan0;

			// Compute the timetotal and delta
			lastTime = curTime;
			curTime = DateTime.Now;
			double timeTotal = (curTime - initTime).TotalSeconds;
			double timeDelta = (curTime - lastTime).TotalSeconds;

			// Run the DLL plugin function
			Form1.pluginStartTime = DateTime.Now;
			Render16bpc(rgb16bpc, timeTotal, timeDelta, vcl);
			Form1.pluginFinishTime = DateTime.Now;

			Form1.pluginAverageTimes.Push((Form1.pluginFinishTime - Form1.pluginStartTime).TotalMilliseconds);

			// Pack the 16 bit channels to 8 bits for display
			Form1.renderStartTime = DateTime.Now;
			RGB16bpcToBMP(A000, rgb16bpc, dithering, vcl);
			Form1.renderFinishTime = DateTime.Now;

			Form1.renderAverageTimes.Push((Form1.renderFinishTime - Form1.renderStartTime).TotalMilliseconds);


			// Unlock the bits
			backBuffer.UnlockBits(bmpData);

			// Set the data pointer back to null
			A000 = null;

			// Clear the surface
			grafx.Graphics.Clear(Color.Black);

			// Draw the image
			grafx.Graphics.DrawImage(backBuffer,
				new Rectangle(0, 0, control.Width, control.Height),
				new Rectangle(0, 0, backBuffer.Width, backBuffer.Height),
				GraphicsUnit.Pixel);

			// Present
			grafx.Render();
		}
	}
}
